@@ -3,13 +3,14 @@ from __future__ import annotations
 
 from collections import defaultdict
 import copy
+from itertools import product
 import logging
 import random
 import re
-from typing import Any, Callable, Dict, List, Optional, Tuple, Union
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union, Set
 
 from .prefetch_generator import PrefetchGenerator
-from ..utils import is_available
+from ..utils import is_available, width
 
 import numpy as np
 
@@ -20,84 +21,159 @@ logger = logging.getLogger(__name__)
 class Dataset:
     """
     """
+    
+    def __init__(self, data: List[Dict[str, Any]] = None, shuffle: bool = False, *args, **kwargs) -> None:
+        if data is None:
+            pass
+        else:
+            if isinstance(data, list):
+                if data:
+                    if isinstance(data[0], dict):
+                        pass
+                    else:
+                        raise(ValueError("각각의 요소는 dict 타입이어야 합니다."))
+            else:
+                raise(ValueError("data는 list 타입이어야 합니다."))
 
-    def __init__(self, data: List[Dict[str, Any]] = None, **kwargs) -> None:
         self._data = data if data is not None else []
-        self._length = 0
+        self._shuffle = shuffle
 
-    def append(self, data: Dict[str, Any]):
+    def append(self, data: Dict[str, Any]) -> None:
         self._data.append(data)
 
     add = append
 
-    def drop(self, condition):
+    def drop(self, condition: function) -> Dataset:
         return self.filter(lambda x: not condition(x))
 
-    def filter(self, condition):
+    def filter(self, condition: function) -> Dataset:
         return Dataset(list(filter(condition, self._data)))
 
-    def merge(self, key, values):
-        new_values = "+".join(values)
-        return Dataset([{k: (new_values if (key == k and d[k] in values) else d[k]) for k in d.keys()} for d in self._data])
+    def merge(self, key: str, values: Union[List[str], Tuple[str]], sort: bool = True) -> Dataset:
+        if isinstance(values, (list, tuple)):
+            if sort:
+                values = sorted([str(v) for v in values])
+            new_values = "+".join(values)
+            return Dataset([{k: (new_values if (key == k and d[k] in values) else d[k]) for k in d.keys()} for d in self._data])
+        else:
+            raise TypeError("unsupported operand type(values) for merge")
 
-    def __add__(self, other: Dataset):
-        cls = Dataset(shuffle=self._shuffle, batch_size=self._batch_size)
+    def __add__(self, other: Dataset) -> Dataset:
+        cls = Dataset(shuffle=self._shuffle)
         cls._data.extend(copy.deepcopy(self._data))
         cls._data.extend(copy.deepcopy(other._data))
-        cls._length = self._length + other._length
         return cls
 
-    def __radd__(self, other: Dataset):
+    def __radd__(self, other: Dataset) -> Dataset:
         return self + other
 
-    def __getitem__(self, index: slice):
+    def __getitem__(self, index: Union[int, slice]) -> Dataset:
         cls = Dataset()
         data = self._data[index]
         if isinstance(data, dict):
             cls._data.append(data)
-            cls._length = 1
         else:
             cls._data.extend(data)
-            cls._length = len(data)
         return cls
 
-    def __len__(self):
+    def __len__(self) -> int:
         return self.num_data
 
     @property
-    def data(self):
+    def data(self) -> List[Dict[str, Any]]:
         if self._shuffle:
             random.shuffle(self._data)
         return self._data
 
     @property
     def num_data(self):
-        if self._length == 0:
-            self._length = len(self._data)
-        return self._length
+        return len(self._data)
 
-    def get_items(self, max_display):
-        _items = defaultdict(lambda: defaultdict(int))
+    def counts(self, keys: Optional[Union[List[str], Tuple[str], Set[str], str]] = None):
+        if keys is not None and isinstance(keys, str):
+            keys = set([keys])
+        _counts = defaultdict(lambda: defaultdict(int))
         for d in self._data:
             for k, v in d.items():
-                if len(_items[k]) < max_display:
-                    _items[k][v] += 1
-        return {k1: dict(v1) for k1, v1 in _items.items()}
-    
-    def print_items(self, max_display=10):
-        _items = self.get_items(max_display)
+                if keys is not None and k in keys:
+                    _counts[k][v] += 1
+                else:
+                    _counts[k][v] += 1
+        return {k1: dict(v1) for k1, v1 in _counts.items()}
+
+    def combination_counts(self, keys: Union[List[str], Tuple[str], Set[str]], views: Optional[Union[List[str], Tuple[str], Set[str], str]] = None) -> List[Dict[str, Any]]:
+        def _cond(keys, values):
+            def _wrapper(data):
+                for k, v in zip(keys, values):
+                    if data[k] == v:
+                        pass
+                    else:
+                        b = False
+                        break
+                else:
+                    b = True
+                return b
+            return _wrapper
+
+        if not isinstance(keys, list):
+            keys = list(keys)
+        
+        if views is not None:
+            if isinstance(views, str):
+                views = [views]
+                
+        _kvs = {k: list(v.keys()) for k, v in self.counts(keys).items() if k in keys}
+
+        combinations = _kvs[keys[0]]
+        for i in range(len(keys)-1):
+            combinations = ["\x1b".join(p) for p in product(combinations, _kvs[keys[i+1]])]
+        combinations = [c.split("\x1b") for c in combinations]
+
+        _counts = []
+        for c in combinations:
+            ds = self.filter(_cond(keys, c))
+            if ds.num_data > 0:
+                _tmp = {k: v for k, v in zip(keys, c)}
+                if views:
+                    _ds_counts = ds.counts()
+                    for v in views:
+                        _tmp[v] = list(set(_ds_counts[v]))
+                _counts.append(_tmp)
+                _tmp["count"] = ds.num_data
+            
+        # max_lengths = [0 for _ in range(len(keys))]
+        # for c in combinations:
+        #     for i in range(len(keys)):
+        #         if max_lengths[i] < width(c[i]):
+        #             max_lengths[i] = width(c[i])
+        # _counts = {}
+        # for c in combinations:
+        #     name = ", ".join([f"{k.title()}: {v:{max_lengths[i]}s}" for i, (k, v) in enumerate(zip(keys, c))])
+        #     n = self.filter(_cond(keys, c)).num_data
+        #     if n > 0:
+        #         _counts[name] = n
+        return _counts
+        
+    def print_counts(self, max_display: int = 10, keys: Optional[Union[List[str], Tuple[str], Set[str], str]] = None, print_all: bool = False):
+        if print_all:
+            max_display = int(1e10)
+        _warn = True
+        _counts = self.counts(keys)
         _repr = []
-        for k, v in _items.items():
+        for k, v in _counts.items():
             _tmp = []
             for _i, (_k, _v) in enumerate(v.items()):
-                _tmp.append(f"{_k if _v == 1 else f'{_k}({_v})'}")
-                if _i+1 == max_display:
+                if _i+1 == max_display+1:
+                    if not print_all and _warn:
+                        logger.warning(f"표시할 요소의 개수가 {max_display:d}개를 초과합니다. 모든 요소를 표시하려면 'print_all=True' 옵션을 사용하세요.")
+                        _warn = False
                     _tmp.append("...")
                     break
+                _tmp.append(f"{_k if _v == 1 else f'{_k}({_v})'}")
             _repr.append(f"{k}: {', '.join(_tmp)}")
         _repr = "\n".join(_repr)
         _repr = re.sub(r"^", " " * 4, _repr, 0, re.M)
-        logger.info(f"Dataset: {self.num_data}\n{_repr}")
+        print(f"Dataset: {self.num_data}\n{_repr}")
 
     def __repr__(self):
         return f"Dataset: {self.num_data}\n"
